#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

/**************************************************************************************************/
volatile uint8_t pulse_ticks=0,prev_pulse_ticks=0x00;
volatile uint8_t status=0x00;

#define stIdle    0x01
#define stAcq     0x02
#define stWait    0x04
#define stSwitch  0x08

#define outMsk    (_BV(PB0)|_BV(PB3)|_BV(PB4))
#define ticksPos0 90
#define outPos0   _BV(PB0)
#define ticksPos1 140
#define outPos1   _BV(PB3)
#define ticksPos2 190
#define outPos2   _BV(PB4)
#define ticksFuzz 12

int main(void)
{
  /* Setup Timer0, in CTC mode
    * SysCLK is 8 MHz,
    * Prescaler = 8, tmrCLK is 1MHz
    * OCR0A is 10, intT is 100kHz (0.01ms)
    * Tmin is 1ms, so ticksMin is 100
    * Tzero is 1.5ms, so ticksZero is 150
    * Tmax is 2ms, so tickMax is 200
    */
  TCCR0A = _BV(WGM01);  /* CTC mode */
  TCCR0B = _BV(CS01);   /* Prescaler /8 */
  TIFR = _BV(OCF1A);    /* Clear OVF flag */
  TCNT0 = 0x00;         /* Reset value */
  OCR0A = 10;           /* TOP value */
  TIMSK = _BV(OCIE0A);  /* Enable interrupt */

  /* Setup INT0 */
  MCUCR |= _BV(ISC01)|_BV(ISC00); /* Rising edge interrupt */
  GIFR = _BV(INTF0);              /* Clear INT flag */
  GIMSK = _BV(INT0);              /* Enable interrupt */

  /* Setup Outputs */
  DDRB = outMsk;
  PORTB = _BV(PB2);               /* Enable pull-up */

  wdt_enable(WDTO_1S);
  sei();
  status = stIdle;
  while(1)
  {
    wdt_reset();
    if(pulse_ticks==0xFF) /* 0xFF => Reset INT0 to rising edge */
    {
      MCUCR |= _BV(ISC00);
      prev_pulse_ticks = pulse_ticks = 0x00;
      status = stIdle;
    }
    if(status & stSwitch)
    {
      if(((prev_pulse_ticks >= pulse_ticks && (prev_pulse_ticks-pulse_ticks)<ticksFuzz) ||
          (prev_pulse_ticks <  pulse_ticks && (pulse_ticks-prev_pulse_ticks)<ticksFuzz)) )
      {
        PORTB &= ~outMsk;
        if(pulse_ticks>ticksPos0-ticksFuzz && pulse_ticks<ticksPos0+ticksFuzz)
        {
          PORTB |= outPos0;
        }
        else if(pulse_ticks>ticksPos1-ticksFuzz && pulse_ticks<ticksPos1+ticksFuzz)
        {
          PORTB |= outPos1;
        }
        else if(pulse_ticks>ticksPos2-ticksFuzz && pulse_ticks<ticksPos2+ticksFuzz)
        {
          PORTB |= outPos2;
        }
        else
        {
        }
      }
      prev_pulse_ticks = 0x00;
      status = stIdle;
    }
    if(status & stWait)
    {
      _delay_ms(100);
      status = stIdle;
    }
  }
}

/* TMR1 ISR **************************************************************************************/
ISR(TIMER0_COMPA_vect)
{
  if((status & stAcq) && (pulse_ticks<0xFF))
    pulse_ticks++;
}

/* INT0 ISR ***************************************************************************************/
ISR(INT0_vect)
{
  if(status & (stIdle|stAcq))
  {
    if(PINB&_BV(PB2))        /* Triggered on Rising edge interrupt */
    {
      pulse_ticks = 0x00;
      status = stAcq;
      MCUCR &= _BV(ISC00);    /* Switch to falling edge */
    }
    else
    {
      if(status & stAcq)
      {
        if(!prev_pulse_ticks)
        {
          prev_pulse_ticks = pulse_ticks;
          status = stWait;
        }
        else
        {
          status = stSwitch;
        }
      }
      MCUCR |= _BV(ISC00);    /* Switch to rising edge */
    }
  }
}

/* Default ISR, do nothing ************************************************************************/
ISR(__vector_default)
{
}
